/**
 * Copyright (c) 2015 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "Tracking/CascadeTracker.h"
#include "Tracking/AsyncTracker.h"

#include "ImageMathUtil.h"
#include <new>

namespace MediaVision {
namespace Image {

CascadeTracker::CascadeTracker(float minimumArea) :
		__trackers(),
		__minimumArea(minimumArea)
{
	; /* NULL */
}

CascadeTracker::CascadeTracker(const CascadeTracker& copy) :
		__trackers(),
		__minimumArea(copy.__minimumArea)
{
	*this = copy;
}

CascadeTracker::~CascadeTracker()
{
	; /* NULL */
}

bool CascadeTracker::track(const cv::Mat& frame, std::vector<cv::Point>& result)
{
	internalReinforcement();

	std::set<TrackerInfo>::iterator it = __trackers.begin();

	for (; it != __trackers.end(); ++it)
		if (!it->mTracker.get()->track(frame, it->mResult))
			it->mResult.clear();

	return mergeResults(result);
}

void CascadeTracker::reinforcement(const std::vector<cv::Point>& location)
{
	std::set<TrackerInfo>::iterator it = __trackers.begin();

	for (; it != __trackers.end(); ++it)
		it->mTracker.get()->reinforcement(location);
}

cv::Ptr<ObjectTracker> CascadeTracker::clone() const
{
	return cv::Ptr<ObjectTracker>(new (std::nothrow)CascadeTracker(*this));
}

CascadeTracker& CascadeTracker::operator=(const CascadeTracker& copy)
{
	if (this != &copy) {
		this->__minimumArea = copy.__minimumArea;
		this->__trackers.clear();

		std::set<TrackerInfo>::iterator it = copy.__trackers.begin();
		for (; it != copy.__trackers.end(); ++it) {
			TrackerInfo temp(it->mTracker.get()->clone(), it->mPriority);
			temp.mResult = it->mResult;

			__trackers.insert(temp);
		}
	}

	return *this;
}

bool CascadeTracker::enableTracker(cv::Ptr<ObjectTracker> tracker, float priority)
{
	TrackerInfo temp(tracker, priority);

	std::set<TrackerInfo>::iterator it =
			std::find(__trackers.begin(), __trackers.end(), temp);

	if (it != __trackers.end())
		__trackers.erase(it);

	return __trackers.insert(temp).second;
}

bool CascadeTracker::disableTracker(cv::Ptr<ObjectTracker> tracker)
{
	TrackerInfo target(tracker, 0);

	std::set<TrackerInfo>::iterator it =
			std::find(__trackers.begin(), __trackers.end(), target);

	if (it == __trackers.end())
		return false;

	__trackers.erase(it);

	return true;
}

void CascadeTracker::internalReinforcement()
{
	std::set<TrackerInfo>::iterator it1 = __trackers.begin();
	for (; it1 != __trackers.end(); ++it1) {
		bool isUpdated = true;

		/* TODO: Redesign without dynamic_cast */
		AsyncTracker *asyncView = dynamic_cast<AsyncTracker*>(it1->mTracker.get());
		if (NULL != asyncView)
			isUpdated = asyncView->isUpdated(it1->mResult);

		if (!it1->mResult.empty() && isUpdated) {
			const size_t numberOfContourPoints = it1->mResult.size();
			std::vector<cv::Point2f> checkedArea(numberOfContourPoints);
			for (size_t i = 0; i < numberOfContourPoints; ++i) {
				checkedArea[i].x = it1->mResult[i].x;
				checkedArea[i].y = it1->mResult[i].y;
			}

			if (getQuadrangleArea(checkedArea.data()) < __minimumArea) {
				it1->mResult = std::vector<cv::Point>(0);
				it1->mTracker.get()->reinforcement(it1->mResult);
			}

			float priority = it1->mPriority;
			std::set<TrackerInfo>::iterator it2 = __trackers.begin();

			for (; it2 != __trackers.end(); ++it2)
				if (it1 != it2 && priority > it2->mPriority)
					 it2->mTracker.get()->reinforcement(it1->mResult);
		}
	}
}

bool CascadeTracker::mergeResults(std::vector<cv::Point>& result) const
{
	result.clear();

	std::set<TrackerInfo>::iterator it = __trackers.begin();

	float resPriotiry = 0.f;
	for (; it != __trackers.end(); ++it) {
		if (result.empty() || resPriotiry > it->mPriority) {
			resPriotiry = it->mPriority;
			result = it->mResult;
		}
	}

	return !(result.empty());
}

CascadeTracker::TrackerInfo::TrackerInfo(cv::Ptr<ObjectTracker> tracker, float priority) :
		mTracker(tracker),
		mPriority(priority),
		mResult()
{
	; /* NULL */
}

bool CascadeTracker::TrackerInfo::operator<(const TrackerInfo& second) const
{
	return (this->mPriority < second.mPriority);
}

bool CascadeTracker::TrackerInfo::operator==(const TrackerInfo& second) const
{
	return (this->mTracker == second.mTracker);
}

bool CascadeTracker::TrackerInfo::operator!=(const TrackerInfo& second) const
{
	return !(*this == second);
}

} /* Image */
} /* MediaVision */
